VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "cBSTNStorage"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = False
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
' BUG FIX   :
' Version   : 1.0 to 1.1
' Details   :  Add optional lpdataindex to allow tree rebalance to keep the
'              new root key dataindex else rebalanced key record data index is lost
'              causing orphaned or mixed up data record, if data index
'              being used to referenc external data records.
'
' Other changes:
'               tidy up nomenclature for parameters passed
'~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
'
' NEWS:  April 2005
'        There is a beta media library sample in development using this class as it's database engine.
'        Other p-s-c visitors offered their help in testing the beta sample, checking the adds, removes,
'        ordered lists etc. and it's overall general usage.
'        The sample maintains details of media such as CDs, DVDs, Audio Tapes, VHS Videos etc.
'        Once testing and functionality has been completed, the sample will be published on p-s-c.
'        Anticipated to be published by middle of May 2005 (or sooner!)
'
'---------------------------------------------------------------------------------------
' Project   : cBSTNStorage
' DateTime  : 25/04/2005 20:37
' Author    : D K Richmond
' Status    : BETA RELEASE 1.1
' Version   : Visual Basic 6 (SP6)
' User Type : Beginner/Intermediate/Advanced
' Applicable: Database/Data Sorting/Key Retrieval/Binary Trees/Node Traversals
'
' Disclaimer:
' -----------
' This software is provided as-is, no liability accepted in using
' this code or part of this code where information is used and relied upon in any
' other system or systems, projects, modules etc.
' In other words, it is provided for educational value and where experimenting
' in writing applications needing fast ordered key retrieval for testing and
' running experimental non-critical non-business applications.
'
' Copyright:
' -----------
' This was written by myself based on information from various Data Abstraction
' and storage techniques.  Some of the deletion routines are based on information (only information not code)
' from various pages on the internet (non-copyright).
' There are some excellent articles on the internet, try google.com .. searching for "Binary Trees"
'
' ALL OF THIS CODE IN THIS CLASS WAS WRITTEN FROM A BLANK WORKSHEET.  IN OTHER WORDS NONE OF THIS CODE
' WAS STRIPPED, EXTRACTED OR COPIED FROM OTHER PEOPLES WORK.  PLEASE RESPECT THAT AND
' IF YOU USE THIS CODE IN ANY OF YOUR APPLICATIONS/EXAMPLES PLEASE GIVE CREDIT FOR MY WORK.
'
' DAVID K RICHMOND dk.richmond@ntlworld.com
'
' GUIDELINES:
' -----------
'  I have used a test harness (not included) to prove these functions work as I expect them to work.
'  I can not release the test harness, it's not pretty, there are lots of non-modal controls and really
'  it would take me weeks to make user-friendly and provide documentation on.  You will not
'  benefit from having it as it was only used to test randomly as many the possible combinations of trees.
'  As I have spent many many hours (that's more than 2 weeks several hours a day), to test and
'  refine the routines to give the best performance and reliablity, there is little testing left to do.
'  I have deleted many lines of code either because they were non-sensical, confusing or poor.  In some
'  cases where I wanted to experiment with different ideas, where I felt the normal established nodal functions
'  were too complicated or too unreliable or too complex to test for 100% data integrity.
'  The most complicated routines in the whole class were the RemoveNode and the
'  GetPrevNodeInSortedOrder member functions.  I spent many hours refining and checking
'  these along with others as they are interelated and relate to the correct removal of a node
'  with two sub-branches.
'
' OVERVIEW:
' -----------
'   Things to be aware of:
'      All entries into the Binary Tree structure are via the root node unless the member
'      function relies on a previous action and the next action is incremental or
'      decrimental e.g. GetPrevNodeInSortedOrder
'      To start a tree you will need to call InitRootNode passing the string key
'      you want to assign to the root node.  Now ideally this root node needs to be the
'      nearest to the middle of the sorted data as possible, but don't worry too much about
'      this as it won't affect anything expect a very small performance hit on searches, inserts and deletions.
'      Having entered a root node the next thing is to do is add the other nodes (any order), using InsertBSTN.
'      The most balanced tree will be from completeley random data inserted sequentially.
'      If you insert sorted data you will end up with a tree with a very long branch or very long branches.
'      This means that the searching will not be the fastest it could be.  Don't worry
'      if you have sorted data, just set the root node, read the data and then before using
'      it try a RebalanceTree passing the middle value of the sorted data.  The searches
'      will then be working on a reasonably balanced tree, i.e faster.
'   Where is the tree store?
'      There is a linear array, which is allocated up by one index for each node added.
'      As each item in the array is of type TBSTNStore is has the node values to allow Binary Tree access.
'   Why can't I find some things referenced in the project:
'       Generally anything included in this release should be of publication standard, e.g. be typed and error checked
'       correctly, laid out properly and essentially reliable and of course bug free!  This project is changing weekly
'       and a release will include only those things worthy of publication.  If after completing a major change,
'       I need to publish the code and an non-publishable item is critical to the exection/compilation I am forced to include it
'       as is.  Further releases will either include the item to the correct standard or it's dependencies removed and
'       itself removed from the project.  These are going to be small items and I hope those of you that may have already built
'       applications based on earlier versions of this code will be impacted heavily by any of these necessary changes.
'       In the next few months the core functionality should stabilise and changes we will see will be limited to additional
'       member functions.  Things like tree rotation, maybe an extra new style tree re-balance etc.  Maybe even some more
'       everyday practical examples to demonstrate how these functions are used.
'
'
' DEBUGGING:
' ---------
' In Class Initialise :  gbDebug = False   '  change this to true to see lots of debug message boxes
'
' SURPRISES:
'  I have left some msgbox's and some redundant error trapping in some routines, this will be taken out real soon.
'  The are some extra definitions that I have left for the future.

' FUTURE:
' -------
' There is a lot of scope to improve this class.
'
' o     Ensuring the tree is always the same level deep
' o     Auto balancing the tree
' o     Dealing with garbage collection
' o     Checking garbage collection thresholds
' o     Tree rotation techniques
' o     Using the Binary Trees for real life applications/examples
' o     Providing some simple test harnesses to test and
'       demonstrate the member functions of the class and the usage
' o     performance counters/traversal statistics

' I will be updating this class, adding extra member functions which may address the above future items or other
' items more pressing or necessary for using this class in other projects.  I have some projects that will make
' use of this class immediately.  If you have ideas, want help in understanding things I have ommitted from the
' overview, usage etc, let me know on the above email. Remember the Binary Tree principals are on the internet.
' Try google.com and use the phrase "Binary Trees" with quotes. Please try to use the example below and have a
' try using the functions yourself before contacting me, the rewards will be better.  Also if for example
' 50+ people contact me I am going to struggle to get a reply off to everyone and have time to improve the
' class. Otherwise look for the next update and I hope to include some interesting and hopefully practical
' examples of using the class to the MAX!!
'
'
' Key Compares.
' -----------
'  Use CompareKeys to check if keys are less, equal or more than each other.  This is because
'  there is a lexical member value.  This makes each key lexical compared, so it makes more
'  sense to human readability.  If you change the lexical to non-lexical you will get true
'  computer sorted keys and hence if you use this you will inherit the same comparisons.
'  If you don't and then later change it and you don't change to code outside the class
'  you will get data inconsistent results.
'
' DONT:  Try this with non-unique keys.  For example A,A,B,C,D,D,E,E
'--------------------------------------------------------------------
' If you do try adding the same key more than once, YOU WILL GENERATE an ERROR.
' This is deliberate.  About 10+ years ago, I wrote a procedure (in C) to allow duplicate
' keys in a Binary tree.  It took me ages to perfect, it was always needing updates/tweaks
' and many coding changes.  I know these days, I could add this quite easily but it's a
' lot more testing! So it has not been included in this release. When I have tested
' using duplicate keys in the same tree and it's reliable and allows correct nodal deletes,
' I will then remove the error trap, make it conditional based on global value and thus allow
' duplicate keys to be stored in the tree.  Of course those brave people reading this
' are welcome to take the class and try with duplicate keys, you don't have to wait for  =[8-)
'---------------------------------------------------------------------------------------
'
' Nomenclature (normal):
' -------------
'      i - integer prefix
'      l - long prefix
'      d - double prefix
'      s - string prefix
'      e - enum
'      c - original class prefix
'      cls - instance of a class prefix
'
'      m_i - member integer prefix
'      m_l - member long prefix
'      m_d - member double prefix
'      m_s - member string prefix
'      m_e - member enum
'      m_cls - member instance of a class prefix
'
' Nomenclature (not normal):
' -------------------------
'      ip - parameter integer prefix
'      lp - parameter long prefix
'      dp - parameter double prefix
'      sp - parameter string prefix
'
'      ipo - parameter optional integer prefix
'      lpo - parameter optional long prefix
'      dpo - parameter optional double prefix
'      spo - parameter optional string prefix
'
'
' NOTES for beginners:
' ---------------------
' The class members are protected from external alteration by being made private.  This protects the integrity
' of the data in the binary tree.   If you end up making the private's into public's you need to re-think why you need
' to do so.   There are public property's to allow Read/Write access where applicable to external programs.
' Please try to keep the private members private and if you need to get the value in a private member
' use a public get property.  Try not to go wild here!
' Private Const are used to protect constant values for being changed during runti
'
' ##########################################################################################################
'
' EXAMPLE USAGE:
'
' Create a new project (standard)
'
' add the following controls:
' o     a listbox named List1
' o     a command button named cmdLoadFile caption "Load File"
' o     a command button named cmdShowSorted caption "Show Sorted List"
'
' then add this code to the form module code
' remember to add both the classes using project menu, add class.
' class:   cBSTNStorage
' and
' class:   cList
'
' now paste this section of example code to the form module code window then
' remove the comment prefix using the ide toolbar function
'
' Run the app, click on the cmdShowSorted button and the letters in the Form_Load
' will appear sorted.
' Now create the file as show in the cmdLoadFile comment for DOS
' click on the cmdLoadFile and see a list of alphabetically sorted filenames
' in the listbox.
'
'  As an example for you to see how easy is it with a binary tree class to find items,
'  write a routine where the item clicked in the list is extracted from the list
'  into a string named sKey. Then use the SearchBSTN passing sKey to located the item in the tree.
'  Use the public property's to retrieve the values of the node Left, Right, Parent and DataIndex.
'
'  Next, I expect having seen some potential, you're thinking, yep, okay, not bad, but what's next?
'  Well, you can use SetDataNodeIndex to assign an index (data type long) to the node.
'  So if you have an array of values elsewhere, you can then assign that index to the node's DataIndex value.
'  Whenever the tree is re-organised the DataIndex value moves with the node.
'  So you can store a whole record, keep the index value and assign it to a key node.
'  You can then perform fast searches and use the GetDataNodeIndex to retrieve the offset to the record.
'
' In SUMMARY: Does that sound similar to a key search on a database record or not?

''  #### start ################### EXAMPLE CODE TO PASTE INTO THE MAIN FORM ##########################
'Dim clsBSTStore As cBSTNStorage
'
'
' Sub Form_Load()
'
'    Set clsBSTStore = New cBSTNStorage
'
'    clsBSTStore.InitRootNode "K"
'
'    clsBSTStore.InsertBSTN "H"
'    clsBSTStore.InsertBSTN "V"
'    clsBSTStore.InsertBSTN "L"
'    clsBSTStore.InsertBSTN "Z"
'    clsBSTStore.InsertBSTN "M"
'    clsBSTStore.InsertBSTN "O"
'
'End Sub
'
'
'Private Sub Form_Unload(Cancel As Integer)
'    Set clsBSTStore = Nothing
'End Sub
'
'
' Sub cmdLoadFile_Click()
'  'use DOS command line:   C:\> dir /b >c:\temp\dir.txt
'    clsBSTStore.InitRootNode "."
'     Open "C:\temp\dir.txt" For Input As #12
'    While Not EOF(12)
'        Line Input #12, sLine
'        clsBSTStore.InsertBSTN sLine
'    Wend
'    Close #12
' End Sub
'
'
' Sub cmdShowSorted_Click()
'
' Dim lNextNode As Long
' Dim sKey As String
'
'    List1.Clear
'
'    clsBSTStore.ResetToRootNode
'
'    Do
'        If clsBSTStore.GetNextNodeInSortedOrder(lNextNode) = True Then
'            Debug.Print String(40, "-")
'            Debug.Print "SO_BTNS[" & lNextNode & "] P=" & clsBSTStore.GetParentNodeOfIndex(lNextNode)
'            Debug.Print "SO_BTNS[" & lNextNode & "] L=" & clsBSTStore.GetLeftNodeOfIndex(lNextNode)
'            Debug.Print "SO_BTNS[" & lNextNode & "] R=" & clsBSTStore.GetRightNodeOfIndex(lNextNode)
'            Debug.Print "SO_BTNS[" & lNextNode & "] K=" & clsBSTStore.GetKeyAtNodeIndex(lNextNode)
'            Debug.Print String(40, "=")
'        Else
'            Exit Do
'        End If
'
'        DoEvents
'
'        If (clsBSTStore.TopOfTreeReached = False) Then
'            sKey = clsBSTStore.GetKeyAtNodeIndex(lNextNode)
'            List1.AddItem sKey
'        End If
'    Loop
'
'End Sub
'
'
''  #### end ################### EXAMPLE CODE TO PASTE INTO THE MAIN FORM ##########################

  
Option Explicit
Dim gbDebug As Boolean

'---------------------------------------------------------------------------------------
' Module    : cBSTNStorage
' DateTime  : 17/04/2005 10:01
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------

Private BSTNStore() As TBSTNStore     ' this is the tree node storage data

Public Enum NodeTypes
    m_ePNODE_TYPE = 1       ' Parent node type
    m_eLNODE_TYPE = 2       ' Left node type
    m_eRNODE_TYPE = 3       ' Right node type
End Enum

' IMPORTANT NOTE:
' if you add any extra items to this type you will need to add them to any routines
' that maintain them during tree re-organisations
Private Type TBSTNStore
    m_lParent As Long       ' Parent Node Link
    m_lLeft As Long         ' Left Node Link
    m_lRight As Long        ' Right Node Link
    m_sNodeKey As String    ' Node Key Value
    m_lNodeDataIndex As Long   ' provides index into external data and this is maintained during tree reorganisations
End Type

' A pretty unique garbage identity key.  If your key value is this then the node has been deleted
Private m_sGarbageIdentityKey  ' = Chr(128) & chr(127) x 5
            
' as it's a percentage of those nodes also allocated on top of store this needs to be reasonably low
Private Const m_cgiGarbageCollectThresholdPercentage = 30

' lexical text matching if set true . ie . a and A are equal value for compares
Private Const m_cbLexical = True
 
Public Enum TraverseOpCodes
    m_eiUNKNOWN = -1
    m_eiGOFARLEFT = 1
    m_eiGOright = 2
    m_eiGOUPVISIT = 3
    m_egiGOUP = 4
    m_eiGOFARRIGHT = 5
    m_eiGOLEFT = 6
    m_eiTOPOFTREE = 99
End Enum

' consts TraverseOpTexts
Private Const m_csGOFARLEFT = "GOFARLEFT"
Private Const m_csGORIGHT = "GORIGHT"
Private Const m_csGOUPVISIT = "GOUPVISIT"
Private Const m_csiGOUP = "GOUP"
Private Const m_csTOPOFTREE = "TOPOFTREE"
' end consts

' tree span counters, watchers
Private m_lNodeHitsCount As Long
Private m_lOffsetLeft As Long
Private m_lOffsetRight As Long
Private m_lNodeLevel As Long
Private m_lCurrentNode As Long
Private m_iOperation As Integer
Private m_lRemovedNodesCount As Long

Private Const m_cNoLink = -1

'---------------------------------------------------------------------------------------
' Procedure : GetDataNodeIndex
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Function GetDataNodeIndex(ByVal lpBSTNNodeIndex As Long) As Long

   On Error GoTo GetDataNodeIndex_Error

    If (lpBSTNNodeIndex >= 0) And (lpBSTNNodeIndex <= GetNodesMaxIndex()) Then
        GetDataNodeIndex = BSTNStore(lpBSTNNodeIndex).m_lNodeDataIndex
    Else
        GetDataNodeIndex = -1
        Err.Raise vbObjectError + 39049, "", ""
        Exit Function
    End If

   On Error GoTo 0
   Exit Function

GetDataNodeIndex_Error:

    MsgBox "Error " & Err.Number & " (" & Err.Description & ") in procedure GetDataNodeIndex of Class Module cBSTNStorage"

End Function

'---------------------------------------------------------------------------------------
' Procedure : SetDataNodeIndex
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Function SetDataNodeIndex(ByVal lpBSTNNodeIndex As Long, ByVal lpObjectNodeIndex As Long) As Boolean

SetDataNodeIndex = False
    
    On Error GoTo SetDataNodeIndex_Error

    If (lpBSTNNodeIndex >= 0) And (lpBSTNNodeIndex <= GetNodesMaxIndex()) Then
        BSTNStore(lpBSTNNodeIndex).m_lNodeDataIndex = lpObjectNodeIndex
    Else
        Err.Raise vbObjectError + 39049, "", ""
        Exit Function
    End If

SetDataNodeIndex = True

   On Error GoTo 0
   Exit Function

SetDataNodeIndex_Error:

    MsgBox "Error " & Err.Number & " (" & Err.Description & ") in procedure SetDataNodeIndex of Class Module cBSTNStorage"
    
End Function

'---------------------------------------------------------------------------------------
' Procedure : GetGarbageCollectThresholdPercentage
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Property Get GetGarbageCollectThresholdPercentage() As Integer
    GetGarbageCollectThresholdPercentage = m_cgiGarbageCollectThresholdPercentage
End Property

'---------------------------------------------------------------------------------------
' Procedure : GetTraverseOperationCode
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Property Get GetTraverseOperationCode() As Integer
    GetTraverseOperationCode = m_iOperation
End Property

'---------------------------------------------------------------------------------------
' Procedure : GetTraverseOperationText
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Property Get GetTraverseOperationText() As String
    GetTraverseOperationText = GetTraverseTextForOpcode(m_iOperation)
End Property

'---------------------------------------------------------------------------------------
' Procedure : GetTraverseCurrentNodeIndex
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Property Get GetTraverseCurrentNodeIndex() As Long
    GetTraverseCurrentNodeIndex = m_lCurrentNode
End Property

'---------------------------------------------------------------------------------------
' Procedure : GetTraverseCurrentNodeKey
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Property Get GetTraverseCurrentNodeKey() As Long
    GetTraverseCurrentNodeKey = BSTNStore(m_lCurrentNode).m_sNodeKey
End Property

'---------------------------------------------------------------------------------------
' Procedure : GetTraverseNodeLevel
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Property Get GetTraverseNodeLevel() As Long
    GetTraverseNodeLevel = m_lNodeLevel
End Property

'---------------------------------------------------------------------------------------
' Procedure : GetTraverseOffsetLeft
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Property Get GetTraverseOffsetLeft() As Long
    GetTraverseOffsetLeft = m_lOffsetLeft
End Property

'---------------------------------------------------------------------------------------
' Procedure : GetTraverseOffsetRight
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Property Get GetTraverseOffsetRight() As Long
    GetTraverseOffsetRight = m_lOffsetRight
End Property

'---------------------------------------------------------------------------------------
' Procedure : TopOfTreeReached
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Function TopOfTreeReached() As Boolean
    TopOfTreeReached = (m_iOperation = m_eiTOPOFTREE)
End Function

'---------------------------------------------------------------------------------------
' Procedure : GetCurrentNodeIndex
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Property Get GetCurrentNodeIndex() As Long
    GetCurrentNodeIndex = m_lCurrentNode
End Property


'---------------------------------------------------------------------------------------
' Procedure : GetKeyAtNodeIndex
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Function GetKeyAtNodeIndex(ByVal lpNodeIndex As Long) As String
    
   On Error GoTo GetKeyAtNodeIndex_Error

    If (lpNodeIndex >= 0) And (lpNodeIndex <= GetNodesMaxIndex()) Then
        GetKeyAtNodeIndex = BSTNStore(lpNodeIndex).m_sNodeKey
    Else
        Err.Raise vbObjectError + 39009, "GetKeyAtNodeIndex", "GetKeyAtNodeIndex"
        Exit Function
    End If

   On Error GoTo 0
   Exit Function

GetKeyAtNodeIndex_Error:

    MsgBox "Error " & Err.Number & " (" & Err.Description & ") in procedure GetKeyAtNodeIndex of Class Module cBSTNStorage"
    
End Function

'---------------------------------------------------------------------------------------
' Procedure : GetLeftNodeOfIndex
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Function GetLeftNodeOfIndex(ByVal lpNodeIndex As Long) As Long

   On Error GoTo GetLeftNodeOfIndex_Error

    If (lpNodeIndex >= 0) And (lpNodeIndex <= GetNodesMaxIndex()) Then
        GetLeftNodeOfIndex = BSTNStore(lpNodeIndex).m_lLeft
    Else
        Err.Raise vbObjectError + 39041, "GetLeftNodeOfIndex", "GetLeftNodeOfIndex"
        Exit Function
    End If

   On Error GoTo 0
   Exit Function

GetLeftNodeOfIndex_Error:

    MsgBox "Error " & Err.Number & " (" & Err.Description & ") in procedure GetLeftNodeOfIndex of Class Module cBSTNStorage"

End Function

'---------------------------------------------------------------------------------------
' Procedure : GetRightNodeOfIndex
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Function GetRightNodeOfIndex(ByVal lpNodeIndex As Long) As Long
    
   On Error GoTo GetRightNodeOfIndex_Error

    If (lpNodeIndex >= 0) And (lpNodeIndex <= GetNodesMaxIndex()) Then
        GetRightNodeOfIndex = BSTNStore(lpNodeIndex).m_lRight
    Else
        Err.Raise vbObjectError + 39049, "GetRightNodeOfIndex", "GetRightNodeOfIndex"
        Exit Function
    End If

   On Error GoTo 0
   Exit Function

GetRightNodeOfIndex_Error:

    MsgBox "Error " & Err.Number & " (" & Err.Description & ") in procedure GetRightNodeOfIndex of Class Module cBSTNStorage"
    
End Function

'---------------------------------------------------------------------------------------
' Procedure : GetParentNodeOfIndex
' DateTime  : 25/04/2005 19:32
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Function GetParentNodeOfIndex(ByVal lpNodeIndex As Long) As Long

   On Error GoTo GetParentNodeOfIndex_Error

    If (lpNodeIndex >= 0) And (lpNodeIndex <= GetNodesMaxIndex()) Then
        GetParentNodeOfIndex = BSTNStore(lpNodeIndex).m_lParent
    Else
        Err.Raise vbObjectError + 39043, "GetParentNodeOfIndex", "GetParentNodeOfIndex"
        Exit Function
    End If

   On Error GoTo 0
   Exit Function

GetParentNodeOfIndex_Error:

    MsgBox "Error " & Err.Number & " (" & Err.Description & ") in procedure GetParentNodeOfIndex of Class Module cBSTNStorage"
    
End Function

'---------------------------------------------------------------------------------------
' Procedure : GetNodesMaxIndex
' DateTime  : 25/04/2005 19:33
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Function GetNodesMaxIndex() As Long
    GetNodesMaxIndex = UBound(BSTNStore())
End Function


'---------------------------------------------------------------------------------------
' Procedure : Clear
' DateTime  : 18/04/2005 08:17
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Function Clear()

   On Error GoTo Clear_Error

    ReDim BSTNStore(0)
        
    BSTNStore(0).m_lParent = m_cNoLink
    BSTNStore(0).m_lLeft = m_cNoLink
    BSTNStore(0).m_lRight = m_cNoLink
    BSTNStore(0).m_sNodeKey = m_sGarbageIdentityKey

    m_lRemovedNodesCount = 0

   On Error GoTo 0
   Exit Function

Clear_Error:

    MsgBox "Error " & Err.Number & " (" & Err.Description & ") in procedure Clear of Class Module cBSTNStorage"

End Function


'---------------------------------------------------------------------------------------
' Procedure : InitRootNode
' DateTime  : 18/04/2005 08:17
' Author    : D K Richmond
' Purpose   :
' BUG FIX   : add optional lpdataindex to allow rebalance to keep the new root key dataindex
'---------------------------------------------------------------------------------------
'
Public Function InitRootNode(ByVal spNodeKey As String, Optional ByVal lpDataIndex As Variant)

   On Error GoTo InitRootNode_Error

    ReDim BSTNStore(0)
        
    BSTNStore(0).m_lParent = m_cNoLink
    BSTNStore(0).m_lLeft = m_cNoLink
    BSTNStore(0).m_lRight = m_cNoLink
    BSTNStore(0).m_sNodeKey = spNodeKey
' BUG FIX   : add optional lpdataindex to allow rebalance to keep the new root key dataindex
    If IsMissing(lpDataIndex) Then
        BSTNStore(0).m_lNodeDataIndex = -1
    Else
        BSTNStore(0).m_lNodeDataIndex = lpDataIndex
    End If
    m_lRemovedNodesCount = 0

   On Error GoTo 0
   Exit Function

InitRootNode_Error:

    MsgBox "Error " & Err.Number & " (" & Err.Description & ") in procedure InitRootNode of Class Module cBSTNStorage"

End Function

'---------------------------------------------------------------------------------------
' Procedure : InitNodeTree
' DateTime  : 25/04/2005 19:33
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
'Public Function InitNodeTree(ByVal spNodeKey As String)
'
'    ReDim BSTNStore(0)
'
'    BSTNStore(0).m_lParent = m_cNoLink
'    BSTNStore(0).m_lLeft = m_cNoLink
'    BSTNStore(0).m_lRight = m_cNoLink
'    BSTNStore(0).m_sNodeKey = ""
'BSTNStore(0).m_lNodeDataIndex = -1
'
'End Function


'---------------------------------------------------------------------------------------
' Procedure : GetTraverseTextForOpcode
' DateTime  : 17/04/2005 10:12
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Function GetTraverseTextForOpcode(ByVal ipTraverseOpCode As Integer) As String
 Dim sOPeration As String
 
    Select Case ipTraverseOpCode
        Case TraverseOpCodes.m_eiGOFARLEFT:
            sOPeration = m_csGOFARLEFT
        Case TraverseOpCodes.m_eiGOright:
            sOPeration = m_csGORIGHT
        Case TraverseOpCodes.m_eiGOUPVISIT:
            sOPeration = m_csGOUPVISIT
        Case TraverseOpCodes.m_egiGOUP:
            sOPeration = m_csiGOUP
        Case TraverseOpCodes.m_eiTOPOFTREE:
            sOPeration = m_csTOPOFTREE
        Case Else
            sOPeration = "?? unknown ??"
     End Select

GetTraverseTextForOpcode = sOPeration
End Function

'---------------------------------------------------------------------------------------
' Procedure : GetNextNodeInSortedOrder
' DateTime  : 17/04/2005 10:01
' Author    : D K Richmond
' Purpose   :
' Notes     : pass thru detects when the node traversal ascends thru a specific node and sets flag
'---------------------------------------------------------------------------------------
'
Public Function GetNextNodeInSortedOrder(ByRef lpNextNode As Long, Optional ByRef lpoCurrentNode As Long = -1, Optional ByRef ipoOperation As Integer = m_eiUNKNOWN, Optional bpoPassThru As Boolean, Optional lpoPassThruNode As Long = -1) As Boolean
 Dim lLeft As Long
 Dim lRight As Long
 Dim lParent As Long
 Dim bMoved As Boolean
 
 GetNextNodeInSortedOrder = False
 
   On Error GoTo GetNextNodeInSortedOrder_Error

    ' @node
    'IF LEFT exist
    '    GO FAR LEFT
    '    next op is go right
    'End If
    '
    'if NOT LEFT exist
    '    if RIGHT exist
    '        GO RIGHT ONE
    '        GO FAR LEFT
    '    End If
    '    ' no left or right
    '    if NOT RIGHT exist
    '        go up
    '        if ascend from right i.e if parent's right is same as current
    '            go up
    '        End If
    '        if ascend from left i.e if parent's left is same as current
    '            at Node
    '        End If
    '    End If
    'End If

    If lpoCurrentNode <> -1 Then
        m_lCurrentNode = lpoCurrentNode
    End If
    If ipoOperation <> m_eiUNKNOWN Then
        m_iOperation = ipoOperation
    End If
    
    bpoPassThru = False
    lpNextNode = m_lCurrentNode
    bMoved = False
    If m_iOperation = TraverseOpCodes.m_eiGOFARLEFT Then
        Do
            lLeft = BSTNStore(lpNextNode).m_lLeft
            If lLeft = m_cNoLink Then
                Exit Do
            Else
                lpNextNode = lLeft
            End If
            If lpoPassThruNode = lpNextNode Then
                bpoPassThru = True
            End If
            m_lNodeLevel = m_lNodeLevel + 1
            m_lOffsetLeft = m_lOffsetLeft + 1
            m_lOffsetRight = m_lOffsetRight - 1
            bMoved = True
        Loop
        If bMoved = True Then
            lRight = BSTNStore(lpNextNode).m_lRight
            If lRight <> m_cNoLink Then
                m_iOperation = TraverseOpCodes.m_eiGOright
            Else
                m_iOperation = TraverseOpCodes.m_egiGOUP
            End If
            m_lNodeLevel = m_lNodeLevel + 1
            m_lCurrentNode = lpNextNode
            GetNextNodeInSortedOrder = bMoved
            Exit Function
        Else
            m_iOperation = TraverseOpCodes.m_eiGOright
            m_lCurrentNode = lpNextNode
            GetNextNodeInSortedOrder = True
            Exit Function
        End If
    End If
    
    'DEBUG
    Debug.Print BSTNStore(m_lCurrentNode).m_sNodeKey
    
    If m_iOperation = TraverseOpCodes.m_eiGOright Then
        lRight = BSTNStore(m_lCurrentNode).m_lRight
        If lRight <> m_cNoLink Then
            lpNextNode = lRight
            If lpoPassThruNode = lpNextNode Then
                bpoPassThru = True
            End If
            bMoved = True
            m_lNodeLevel = m_lNodeLevel + 1
            m_lOffsetRight = m_lOffsetRight + 1
            m_lOffsetLeft = m_lOffsetLeft - 1
            Do  'GOFARLEFT
                lLeft = BSTNStore(lpNextNode).m_lLeft
                If lLeft = m_cNoLink Then
                    Exit Do
                End If
                lpNextNode = lLeft
                If lpoPassThruNode = lpNextNode Then
                    bpoPassThru = True
                End If
                m_lNodeLevel = m_lNodeLevel + 1
                m_lOffsetLeft = m_lOffsetLeft + 1
                m_lOffsetRight = m_lOffsetRight - 1
                bMoved = True
            Loop
            If bMoved = True Then
                lRight = BSTNStore(lpNextNode).m_lRight
                If lRight <> m_cNoLink Then
                    m_iOperation = TraverseOpCodes.m_eiGOright
                Else
                    m_iOperation = TraverseOpCodes.m_egiGOUP
                End If
                m_lCurrentNode = lpNextNode
                GetNextNodeInSortedOrder = bMoved
                Exit Function
            Else
                 m_iOperation = TraverseOpCodes.m_egiGOUP
            End If
            m_lCurrentNode = lpNextNode
            GetNextNodeInSortedOrder = bMoved
            Exit Function
         Else
            m_iOperation = TraverseOpCodes.m_egiGOUP
        End If
     End If
     
    If m_iOperation = TraverseOpCodes.m_egiGOUP Then
        lParent = BSTNStore(m_lCurrentNode).m_lParent
        If lParent <> m_cNoLink Then
            m_lNodeLevel = m_lNodeLevel - 1
            Do
                If BSTNStore(lParent).m_lParent = m_cNoLink Then
                   ' MsgBox "eek!"
                    'Err.Raise vbobjecterror +940945, "", ""
                    'Exit Function
                    ValidateTree  ' DEBUG
                End If
                If BSTNStore(lParent).m_lRight = m_lCurrentNode Then
                    m_lOffsetRight = m_lOffsetRight - 1
                    m_lOffsetLeft = m_lOffsetLeft + 1
                    ' ascending don't show node
                    m_lCurrentNode = lParent
                    If lpoPassThruNode = m_lCurrentNode Then
                        bpoPassThru = True
                    End If
                    lParent = BSTNStore(m_lCurrentNode).m_lParent
                    If lParent = m_cNoLink Then
                        m_iOperation = TraverseOpCodes.m_eiTOPOFTREE
                        lpNextNode = 0
                        m_lCurrentNode = lpNextNode
                        GetNextNodeInSortedOrder = False
                        Exit Function
                    Else
                        m_lNodeLevel = m_lNodeLevel - 1
                    End If
                Else
                    If BSTNStore(lParent).m_lLeft = m_lCurrentNode Then
                        m_lOffsetLeft = m_lOffsetLeft - 1
                        m_lOffsetRight = m_lOffsetRight + 1
                        ' ascending show node
                        m_lCurrentNode = lParent
                        lpNextNode = lParent
                        If lpoPassThruNode = lpNextNode Then
                            bpoPassThru = True
                        End If
                        bMoved = True
                        m_iOperation = TraverseOpCodes.m_eiGOright
                        m_lCurrentNode = lpNextNode
                        GetNextNodeInSortedOrder = bMoved
                        Exit Function
                    End If
                End If
            Loop
        Else
            m_iOperation = TraverseOpCodes.m_eiTOPOFTREE
        End If
    End If

GetNextNodeInSortedOrder = bMoved

   On Error GoTo 0
   Exit Function

GetNextNodeInSortedOrder_Error:

    MsgBox "Error " & Err.Number & " (" & Err.Description & ") in procedure GetNextNodeInSortedOrder of Class Module cBSTNStorage"
End Function


'---------------------------------------------------------------------------------------
' Procedure : GetNextAllocationNode
' DateTime  : 17/04/2005 10:01
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Private Function GetNextAllocationNode() As Long
    
    GetNextAllocationNode = UBound(BSTNStore) + 1
    
End Function


'---------------------------------------------------------------------------------------
' Procedure : SearchBSTN
' DateTime  : 17/04/2005 10:01
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Function SearchBSTN(ByVal spKey As String, ByRef lpVisitedNode As Long, Optional ByVal lpCurrentNode As Long = 0) As Boolean
 Dim bNodeFound As Boolean
 Dim lLeft As Long
 Dim lRight As Long
 Dim sKey As String
 
    m_lNodeHitsCount = m_lNodeHitsCount + 1
 
    bNodeFound = False  ' default result is not found
    
    If lpCurrentNode = m_cNoLink Then
        SearchBSTN = False
        Exit Function
    End If
    
    lpVisitedNode = lpCurrentNode
    
    sKey = BSTNStore(lpCurrentNode).m_sNodeKey
    
    If m_cbLexical = True Then
        sKey = UCase(sKey)
        spKey = UCase(spKey)
    End If
    
    If spKey = sKey Then
        ' yes key found
        bNodeFound = True
    End If
    If spKey < sKey Then
        ' check left branch
        lLeft = BSTNStore(lpCurrentNode).m_lLeft
        If lLeft <> m_cNoLink Then
            bNodeFound = SearchBSTN(spKey, lpVisitedNode, lLeft)
        Else
            Exit Function
        End If
    End If
    If spKey > sKey Then
        ' check right branch
        lRight = BSTNStore(lpCurrentNode).m_lRight
        If lRight <> m_cNoLink Then
            bNodeFound = SearchBSTN(spKey, lpVisitedNode, lRight)
        Else
            Exit Function
        End If
    End If

SearchBSTN = bNodeFound
End Function


'---------------------------------------------------------------------------------------
' Procedure : Class_Initialize
' DateTime  : 17/04/2005 10:18
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Private Sub Class_Initialize()

    gbDebug = False   '  change this to true to see lots of debug message boxes
    
    m_sGarbageIdentityKey = String(5, Chr(128) & Chr(127))
    
    InitRootNode m_sGarbageIdentityKey
    
    m_lRemovedNodesCount = 0
    
End Sub


Public Function ResetToRootNode() As Boolean

    ' reset counters, trackers.
    m_lOffsetLeft = 0
    m_lOffsetRight = 0
    m_lNodeLevel = 0
    
    ' reset traversal params
    m_lCurrentNode = 0
    m_iOperation = TraverseOpCodes.m_eiGOFARLEFT

ResetToRootNode = True
End Function


'---------------------------------------------------------------------------------------
' Procedure : InsertBSTN
' DateTime  : 17/04/2005 11:27
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
'
Public Function InsertBSTN(ByVal spKey As String, Optional ByRef plAllocatedNode) As Boolean
 Dim sKey As String
 Dim bNodeFound As Boolean
 Dim lVisitedNode As Long
 'Dim lAllocatedNode As Long
 Dim sVisited As String
 Dim sAllocated As String
 
    InsertBSTN = False
    plAllocatedNode = -1
    m_lNodeHitsCount = 0
 
    If BSTNStore(0).m_sNodeKey = m_sGarbageIdentityKey Then
        MsgBox "Tree is empty going to replace root node with " & spKey
        BSTNStore(0).m_sNodeKey = spKey
        InsertBSTN = True
        plAllocatedNode = 0
        Exit Function
    End If
    
    bNodeFound = SearchBSTN(spKey, lVisitedNode)

    ' prevent duplicates for now
    If bNodeFound = False Then
        plAllocatedNode = GetNextAllocationNode()
        
        ReDim Preserve BSTNStore(plAllocatedNode)
        
        BSTNStore(plAllocatedNode).m_sNodeKey = spKey
        
        sAllocated = BSTNStore(plAllocatedNode).m_sNodeKey
        sVisited = BSTNStore(lVisitedNode).m_sNodeKey
        If m_cbLexical = True Then
            sAllocated = UCase(sAllocated)
            sVisited = UCase(sVisited)
        End If

        ' traversals must end on a leaf of some kind so this is valid
        ' no checking of m_cNoLink required.
        If sAllocated < sVisited Then
            BSTNStore(lVisitedNode).m_lLeft = plAllocatedNode
        Else
            BSTNStore(lVisitedNode).m_lRight = plAllocatedNode
        End If
        
        BSTNStore(plAllocatedNode).m_lParent = lVisitedNode
        BSTNStore(plAllocatedNode).m_lLeft = m_cNoLink
        BSTNStore(plAllocatedNode).m_lRight = m_cNoLink
        BSTNStore(plAllocatedNode).m_lNodeDataIndex = m_cNoLink
        InsertBSTN = True

    End If
    
End Function


'---------------------------------------------------------------------------------------
' Procedure : ReleaseRootNode
' DateTime  : 17/04/2005 13:37
' Author    : D K Richmond
' Purpose   : used to release the root node prior to node removal
'---------------------------------------------------------------------------------------
'
Public Function ReleaseRootNode() As Boolean
 Dim ListX As cList
 Dim lNextNode As Long
 Dim iOperation As Integer
 Dim sKey As String
 Dim sOPeration As String
 Dim lIndex As Long
 Dim lMax As Long
 Dim lLeft As Long
 Dim lRight As Long
 Dim lParent As Long
 Dim sRebalanceKey As String
 Dim sRootNodeKey As String
 Dim iRandom As Integer
 Dim bOk As Boolean
 Dim lDataIndex As Long
 
ReleaseRootNode = False
 
    Set ListX = New cList
    
    ' use list to rebuild keys held in the tree one side of root node key
    ListX.Clear
 
    lParent = GetParentNodeOfIndex(0)
    lLeft = GetLeftNodeOfIndex(0)
    lRight = GetRightNodeOfIndex(0)
    sRootNodeKey = GetKeyAtNodeIndex(0)
    sRebalanceKey = ""  ' if fails nothing happens

    If (lLeft = m_cNoLink) And (lRight <> m_cNoLink) Then
        ' rebalance on right
        sRebalanceKey = GetKeyAtNodeIndex(lRight)
        lDataIndex = GetDataNodeIndex(lRight)
    End If
    If (lLeft <> m_cNoLink) And (lRight = m_cNoLink) Then
        ' rebalance on left
        sRebalanceKey = GetKeyAtNodeIndex(lLeft)
        lDataIndex = GetDataNodeIndex(lLeft)
    End If
    If (lLeft <> m_cNoLink) And (lRight <> m_cNoLink) Then
        iRandom = Int(Rnd(1) * 2)
        If (iRandom \ 1) = 1 Then
            sRebalanceKey = GetKeyAtNodeIndex(lLeft)
            lDataIndex = GetDataNodeIndex(lLeft)
        Else
            sRebalanceKey = GetKeyAtNodeIndex(lRight)
            lDataIndex = GetDataNodeIndex(lRight)
        End If
    End If
    
    bOk = RebalanceTree(sRebalanceKey, False, lDataIndex)
    
    If bOk = False Then
        MsgBox "Problem with Rebalance in ReleaseRoot Node", vbCritical
        Err.Raise vbObjectError + 4054095, "", ""
    End If
    
ReleaseRootNode = bOk
End Function

'---------------------------------------------------------------------------------------
' Procedure : WhatGarbage
' DateTime  : 17/04/2005 20:31
' Author    : D K Richmond
' Purpose   :
'---------------------------------------------------------------------------------------
' notes: 'there is a time penalty for each time to call this if dynamic used
Public Function WhatGarbage(ByVal bAutoGarbageCollect As Boolean, Optional ByVal bDynamic As Boolean = False) As Double
Dim lMax As Long
Dim lIndex As Long
Dim lCount As Long
Dim dPercentGarbage As Double
Dim bOk As Boolean

    lMax = GetNodesMaxIndex()
    If bDynamic = True Then
        lCount = 0
        For lIndex = 0 To lMax
            If GetKeyAtNodeIndex(lIndex) = m_sGarbageIdentityKey Then
                lCount = lCount + 1
            End If
        Next
    Else
        lCount = m_lRemovedNodesCount
    End If
    
    If (lCount > 0) And (lMax > 0) Then
        dPercentGarbage = lCount / lMax * 100
    Else
        dPercentGarbage = 0
    End If
    
    If dPercentGarbage > m_cgiGarbageCollectThresholdPercentage Then
        If bAutoGarbageCollect Then
            MsgBox "Garbage Collection Activated G%=" & dPercentGarbage & " >  GMAX%=" & m_cgiGarbageCollectThresholdPercentage
            bOk = RebalanceTree()
            If bOk = False Then
                MsgBox "Garbage Collection Failed!"
                Err.Raise vbObjectError + 3490394, "", ""
            End If
        End If
    End If

WhatGarbage = dPercentGarbage
End Function

'---------------------------------------------------------------------------------------
' Procedure : RebalanceRootNode
' DateTime  : 17/04/2005 20:36
' Author    : D K Richmond
' Purpose   : Used to rebalance the tree around a central key value - may promote a need
'               to call Garbage Collection
' Notes     : uses pseudo non-gui list class
'
' BUG FIX   : add optional lpdataindex to allow rebalance to keep the new root key dataindex
' NOTE: if around middle is used the local lpdataindex is recalculated else the data index passed is used
'         for the new root node
'---------------------------------------------------------------------------------------
'
Public Function RebalanceTree(Optional ByVal spKey As String, Optional ByVal bpoAroundMiddle As Boolean = False, Optional lpoDataIndex As Long) As Boolean
 Dim lAllocatedNode As Long
 Dim lNextNode As Long
 Dim sKey As String
 Dim lIndex As Long
 Dim lMax As Long
 Dim sRebalanceKey As String
 Dim lRebalanceDataIndex As Long
 Dim sNodeKey As String
 Dim lNodeCount As Long
 Dim lNodesMiddle As Long
 Dim List4 As cList
 Dim List6 As cList
 Dim bOk As Boolean

RebalanceTree = False

    If spKey = "" Then
        sNodeKey = GetKeyAtNodeIndex(0)
    Else
        sNodeKey = spKey
    End If
    
    If sNodeKey <> m_sGarbageIdentityKey Then
        sRebalanceKey = sNodeKey
        lRebalanceDataIndex = lpoDataIndex
    Else
        MsgBox "Tree is empty!", vbCritical
        Exit Function
    End If
    
    Set List4 = New cList
    Set List6 = New cList
    ' use list to rebuild keys held in the tree one side of root node key
    List4.Clear
    List6.Clear
    
    lNodeCount = 0
    lNodesMiddle = CLng((GetNodesMaxIndex() + 1) / 2)
    ResetToRootNode
    Do
        If GetNextNodeInSortedOrder(lNextNode) = True Then
            Debug.Print String(40, "-")
            Debug.Print "SO_BSTN[" & lNextNode & "] P=" & GetParentNodeOfIndex(lNextNode)
            Debug.Print "SO_BSTN[" & lNextNode & "] L=" & GetLeftNodeOfIndex(lNextNode)
            Debug.Print "SO_BSTN[" & lNextNode & "] R=" & GetRightNodeOfIndex(lNextNode)
            Debug.Print "SO_BSTN[" & lNextNode & "] C**=" & GetKeyAtNodeIndex(lNextNode)
            Debug.Print String(40, "=")
        Else
            Exit Do
        End If
        If (TopOfTreeReached = False) Then
            sKey = GetKeyAtNodeIndex(lNextNode)
            List4.AddItem sKey
            List6.AddItem BSTNStore(lNextNode).m_lNodeDataIndex
            lNodeCount = lNodeCount + 1
            If bpoAroundMiddle = True Then
                If lNodeCount >= lNodesMiddle Then
                    sRebalanceKey = sKey
                    lRebalanceDataIndex = BSTNStore(lNextNode).m_lNodeDataIndex
                    bpoAroundMiddle = False
                End If
            End If
        End If
        If (TopOfTreeReached = True) Then
            Exit Do
        End If
    Loop
    
    Debug.Print String(40, "*")
    
    lMax = List4.ListCount - 1
    For lIndex = 0 To lMax
        Debug.Print "SO_BSTN[" & lIndex & "] C**=" & GetKeyAtNodeIndex(lIndex)
    Next
    Debug.Print String(40, "=")
    Debug.Print "SO_BSTN Root [" & 0 & "] C**=" & sNodeKey
    
    If sRebalanceKey = "" Then
        InitRootNode m_sGarbageIdentityKey
        MsgBox "Tree is empty", vbPe Then
            Oew roe--------------------------odeOfIn
      ----------e   lMaN Root [" & 0 & "] C**_lCurrentNode).    EndBSTN[" &  (len
            Oew r L="               Ude). Author   BsGarbageIdennch
        lRight = BSTNStore(lpCuGarbagx "Tree ise           Ude). Auth>or    : D K Richmond
' Purpose   : Use             En------CCCCCCCCCCCCCCCCC------1rpose  4ight = OO_BS& "] CelpNodealParen3     ----CC0anceKey = sKey
                    lRebalanceDaed < TBSTNStoound
        bN D K Richmond
" & modesCount = 0

   On" & co B.ciodesCouStoound
 (oound
 (oound
 (oos:(oos:.m_lRight = m_cNoLink
        BSTNStoro------ : InitNo_lRight =            DebuSe ise >or            sRebalanceKeyDgtrx "SO="       NoPass    
        InitRootNode m_sGarbageIdentityKey
        MsgBox "Tree is empty", vbPe Then
            Oew roe--nSortedOrder(LLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLLL)GoToode       : InitNo_l             f0GetLeftN t6.Agtrx "SO="        -----ed)yoLLLL)GoToode      D0 trx "SO="     fring
 Dim iRandowLL)GoToode      D0IRim lAllocatedNode As Long
 o sKey
True) Then
            Exit DoLLLLLLLLLLLLLLLLLLLLLLLLLLLLLExit Do
        End IfStore(((((((((((((((((((((((((((ring plAllocatedNodeode &n thend Ilright
        sRe(((ring plAllocGarbaex(lNextNode)
            Debug.Print "SO_BSTN["pNodeI.&
    Set List6 = -Idennch
        y As St    f0tus_InitoundMiddle = FaS  GetNext-Idennch
 ""  ' if fails noth
        y As St    f0tus_InitoundMiddle = FaS  GetNext-Idennch
 ""  '       DebuSe ise >or            sR -Idennh
  FaS  GetNext-Idenn       l.odeLevel - 1
         ey = "" Then
        InitRoot
     f------------------- Oew r L=LLLLLLLLLLLLLLLLLLLStore" Then
        InitR Richmond
' 
 :nitRoise >oionim s
 :nitRoise >oionim s
 :nitRous_InitoundMiddl [" & 0 & "] C**=" & sNodeKey
    
    If sRebalanceKey = "" Then
        InitRootNode m_sGarbageIdentityKey
        MsgBoxNext-Idenn       l.odeLevel - 1
         ey = "" Thionim s
 :nitRous_InitoundMiddl [" & 0 & "] C**=" & sNo 4ight = OO_BS&re : InitNodeTree
' DateTime  : 25/04/2
        lRes ----rt                      m_lCurrentNode = lpNextNode
                        GetNextNodeInSortedOrder o(- tlist class
'
'Y tNodeIndex(lNextNode)     lRx(lNextNode)
 NextNode
                        GetNextNodeInuim  lRx(letNextNodeInuim  lRx(le lParent 
   e ise >or :uim  lRx(letNextNodeInuoR(TCritical
         oound
 (oound
 (oos:(oos            GetNextNodeInu     = FaS  GetNext-Idenncx(lNextNode)
  SearchBS   BSTNStod
  (len
            Oew r L="               Ude). Author   BsGarbageIdennch1'locGarbaex(lNextNode)
 r L=        GetNextNodeInuE ) Then
            Exit Do
LLLLLLLLLLLLLLLLLLLLLLLExit Do
        End IfStore(((((((((((((((((((((((((((ring plAlloNtmarbageI
  FatllocaOkNoLT & Chr(127))
=F?         Oew r L="               Ude). Author  arbageantNode)
 r L=        GetNextNodeInTation If (iR     Nob=                       InitRootNodeext-IddddTation If (iR     Nob=                       InitRootNodeext-IddddTation If (iR   onal ByVal bpoAroundMiddle As Boolean = False, Optional lpoDataIndex As Long) As Boolean
 Dim lAllocatedNode As Lon    I((((((((((((ring plAlloNtmarbageI
  FatllocaOkNoLT & Chr(127))
=F?         Oew r L="               Ude). AuthorrentoData_InitoundMic1) = 1 Then
            sRebalanceKey = GetKeyAtNodeIndex(lLeft)
            lDataIndex = GetDataNodeInde AuthorrrrraIndionim s
 :nitRous        Ude). AutnitRouDsAy'    -----ed)yoLLLL)aRous        Ude). AutnitRouDsAy'    -----ed)yoLLLL)re(lVisitedNode).m
t-----odeOfIn
      ------          GetNextNoe = False

    If spKey = "" Then
        sNodnWe

    If spKey = "" Then
        sNodnWe

R"Tree is empty", vbPe Then
     lDat  Els   Loo
     lDat  Els   Loo
     lDat  Els
        MsgBox "Trt-----odeO
    bNowg     --->        eKey = GetKeyAtNodeIndex(lLeft)
         O  Els   Loo
     lDy
    
    If sRebalanceKey M3 
        pty!", vn = N(lVistKeyA           m_lNodeLevel = m_L  bNowg      TN(lVistKeyA     =           8
'
'Y tLLLLLLLLLLLLLLLLLLLLLLLLLLLLL)GoToode       : InitNo_l    m_nitRLLLLS)e                              sRe   l       pi= m_lOffsetLeft + 1111111"eft + 1111NoLT & Chr(12RmtNodeInuim  lRx(le lParent 
   e ise >>>>eyblRx(le lParent 
   e =      GetNextNohruNarent 
   e =     
         is eftNLLLLLLL- etNextNohruNarent 
   e =  l_r  Els   Loo
  en9         m_lNodeLevel = m_L   en9         m_lNsst4.AddItem sKey
            List6.AddItem BSTNStorLLLLLLLLLLLLLLLLLLLLLLLLLLLL  GO FAR LEFT
    ndMi     etRight - 1
              ocatedNode As Long
 Dim sVisited As String
 Dim sAllCey As TNS  GO FAR LEe

R". AutnitRouDsAy'    n
wLL)GoToode      D0IRim lAlllass
ANode = 0
        Exi' DateTime  : 25/R_lNodeLevel = Rim lAlllass
ANode = 0
re(lNextNode).m_lNodeDataRP
   SearchBSTN = bNodeFound
End F      p        E.l1t      Rn_dylAllocate----e   lMaN Root [" & eN(lVistKeyA   = bNodeFound
ENodeFoundlVistK8'     If lpoPassThruNode s       ')GoToode       : InitNo FAR LEe<> m_cNoLink) And (lRight <> m_cNoLink) Then
        iRLLL)aRous        Ude).ed If
ls   L0                    m_lCurrentNode 2urrentNode 2ur+ 1tllocaOkNoLT & Chr(127))(len
)or    : D K u' Author   Getumond
' Purpose   :
'--------------------------- : Clear
' Datehl?text-IdTN = bNodeFus_InitoundMiddle = FaS  GetNext-Idenn----      GetNext-Id_InitoundMiddle = FaS  GetNextnd (lde 2-------------tects when the node s Stri    alance in ReleaseRom_sGarbageLLL)GoTooB [" & 0 & "]tNexxxxxxxxxxxxxxxxxxxxxxxxxx        orbageLLL)GoTooB [" & 0 & "]tNexxxxxxxxxxxxxxxxxxxxxxxxxx     ?"]tNexxxxxx
   elxxxxLLLLL)GoToode       : I9tos LonNodeCount As LtedNode) xxxxxxxxxxx        orbageLLL)GoTooB [" & 0 & "NNNNNNNNNNNN    orbageLLL)GoTooB ["leasle lParent 
    NNN              m_iOperation = Traversea : InitNo FAR LEe<> m_cNoLink) A GoTooB ["leasKey = "O/hitNo FGoToo         geLLL)GoTooB ["leasle lParent 
    NNN  P(lRi       Ude). AutitNo FGoToo         geLLL)GoToo,)
=AoTooB [" & 0 & "]tNexxxxxxxxxxxxxxxxxxxxxxxxxx   -EhlAllocatedNodeode &n thk) AtGarbaCodesbat LLL)GoTooB ["lu 0 & "]tNexxxxxxxxxxxxxxxxxxxxxxxxxx   -EhlAllocatedNodeodIdenncPublic Functioctioode AsT & Chr(127----ooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooxtNodeInuoR(TCritical
         ooundoToo,)
=        ooundoToo,)
=      xxxxxxxxxxxxxnFAR LEe<> m_cNoLink) A GoTooB ["leasKey = "O/hidNode As Long, Optionaheld in the tree ou):ni%Xxxxxxx          ")
As Long,tedNode As Long
   M& 0 & ""oooooo m_L   en9 ---
'
Public Function RebalanceTree(Optional ByVal spKey As String, Optional ByVal bpoAroundMiddle As BoolealLeft)
rsea : InitNo FAR Loooooooooooooooooooooooooooooooooooooooooooo: (lde 2e=nd If
    
   OndoTrtNode = lpeAuthor  'OtToRootNode = True
End Function


'-------------------------------------------------------------------------------------------------------------------------------------------------------------------------------------Kgde = lLeft
                If lpoPassThruNode = lpNextNode Then
              ---------------------& "]tNexxxF      Cchmond
54095f"]tNexxtion

l + 1
            =oooole = FaS OptionaycNoLink) A GoTooB ["leas B1
         etion.ex)
    
    nal ByVal spKdentink) Then
      ocate2orbageLLLooooooooooooo    GetKeyAtNyAtNyAtNyAtNyAt= "O/hitNo FGoTo  Lis---------------& "]tNexxxF   CqTNStore(lpCurrentNode).m_lLeft
        If lLe CqTNStore(lpCurreNode)-Em_lLeftk = RebalanceTre0t StPare  
  reNode)-Em_lTootNyApNStore(lpNextNodNStordNN  P(lRi       Ude). Autite LEe<> m_cNoLink) A GoTooB ["leasKey = "O/hidNode As Long, Optionaheld in the tree ou):ni%Xxxxxxx          "Pe  [" & 0 & "NNNNNNNNeeore(0).m_sNodeKey = bpoAroundMiddle = TrIndex"
        dex)
 alanceKey = sKey
'= bpoAroundMiddle = TrIn'' Auth4 Chr(1TNStore(lg(40, "*")
    
    lMax = Lis  eKeooB ["leasKey =-----------------Kgde = lL------ m_cNoLink) A GoTooBg-----------------------Kgde = lLeft
                                                 LLLL  Gd As    
    lMax = Lisentink) Then
   = FaS OptioooooooxtNodeInuoR(TCritical
         ooxxxF ICit Funt LLL)eore(0Key
        MsgBoxNext-IdFunction RebalanceT9---Kgdxt-IdFunction RebalanceT9---Kgdxte  "lpNextNode  
 S + 111el  ooNodeLetion = Traversea : InitNo FAR LEe<> m_cNoLink) A GoTooB ["leasKey = "O/hitNo FGoToo         geLLL)GoTooBLong, Optionaheld 
       m_cNoLiL  Gd As,nitNo FAlanceT9---Kgdxt-IdFu
ANode = 0
      ule cBSode YL----- P(lR'
Public Function ReR_lNodeLeg
   BInitRootNodeext-IddddTation If (iR   onal ByVal bpoAOaCo9a0eR_lNod0
     = bNodeFound
End F
     bat L= FaS  GetNext-Iden    = bNodeFound
Next-Iden    = bNodeFo
Next-Iden    = bNodeFo
Nex
    :Cunt >= lNodesMid0
    :Cunt >= lNodeeToo                  lNodee<> Cunt  ["leas B1
    ------          "Pe  [" dNStordNNR_lNodT-IdFu
ANode = (((((((ring pl
   ----------eToo   lLe tPare  
  reNode)-Em_lTootNyApNStore(lpNextNodNStordNN  P(lRi       Ude). Autite LEe<> m_cNoLi = sKey
'= bpoAroun "]tNexxxxxxxStB ["leasKey =nk) Then
   = FFFFFFFFFF= sKey
'= bpoArounodIdenncPublic
 Dim sA m_l0
      ule cBSode  = FFFFFFFFFF= sKey
'= bpoArounsC
         O  Els   Loo
     lDy
    
    If sRebalanceKey M3 
        pty!=dIdenncPublic Functioctiood in = Lisentink) Then
   = FaS OptiByVal spKdentiCunt  ["
'=x = Lis  N       geLLL)GoToCN       geLLL)GoToCNlpNeTlic
 Dim sA m_l0
 n
'= Dim sA m_l0
 n
'=itRootNodMMMMM
        Case = Rebl  ElsMMM
      ---------rn    = bNodeFo
Next-Iden    = ElsMMM
      ---------rs   Loo
     lDat  El rLpKdenEl rLpKdeL0
        Exiat  R-----eFo
Next-Iden    = ElsMMM
      -s key founoooooooo: (lde 2e=nd If
    
   O = bN: (ld      c
    
   O = bN: (  m_c0000000:Cunt >= I1
      Exi' DateTi--------eTlRight)
        End If
AGetKey=AtNyAt= "O/hitNo FGoTo  Lis--------------m-te Thenb                 End If
                ule cBSode  =
 Dim s'MDaLL)GoToCNlpNeTlsea : InitNo FAR Lh[" & 0 & "]tNexxxxxxxxxxxxxxxxxxxxxxxxxx   -EhlAllocatedNodeode &n thk) AtGarbaCodesbat LLL)GoTooB FGof
            End If
        End If
    0_lNsst4.Adl
'---------------------------RebalanceT9---Kgdxt-I      ssssssssssssssssssssssssssssssssssssssssK,Nex
    :Cunt >= /hidNode ssssssssssssssssK,Nex
    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu    :Cu  ---------e   lMaN Root [k.    0_lNsst4.Adl
'---------------------------Reb ----       sReb-     n
'=itRootNodMMMLrNN  P(lRieu    :Cu    :Cu    :Cu    :Cu    :Cu      uleeKey = "" Then
     N'    -----ed)yoLLLL)aRous        Ude). AutnitRouDsAy'    -----ed)yoLLLL)re(lVisitedNode).m
t-----odeOfIn
      ------          GetNextNoe = False

    If sO_BSTN[" & lNexN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   IN   INphedrNN  P(lRieu    :Cu    :Cu    :Cu    :Cu    :Cu      uleeKey =LLLLLLLL=
   tNexxxxxx As Long, OptionaheldaN RooLOg, xxxxxxxxxxxxxxxxxxxxxxx   ociStore(lpNe ym lNodeCou= "O/ssssss  :CuLL=
   tNexxx =LLedure : InitNodeTr   I54095f"]tNexxtioUde). Autite LEe<> m_cNoLink) A GoTooBl( L= Fode)     N   IN   I ss  :CuLL=
   tNexxx =LLedure : InitNodeTr   I54095f"]tNexxtioUde). Autite LEe<> m_R   onal ByVal bpR   onal Bhitionaheld l ByeFo
Next-Iden    = ErR  e<> m--------------------------------------LrNN  P=----------------------------LrNN  P=--------movedNodesCount = 0
    
End Sub


Public Function ResetToRootNode() As Boolean

    ' reset counters, trackers.
    m_lOffsetLeft = 0
    m_lOffsetRight = 0
    m_lNodeLevel = 0
    
    ' reset traversa) A   '0000-----------oooooot       Else
     IN   IN   IN   IN   IN   IN  N   IN   IN  N   IN   IN  N   IN   IN  IN   IN   IN   WedNodeode &n thk)--------------igh  ---------rn  deInde AuthorrrrraIndio_xxxxxUSub


Public
E
'= bpoArounodink) A GoTooB------------------------------------------------------------------------------- -----lA : Ini5= bNodeFo
Nex
    :Cunt >= lNoB deInde AuthorrrrraIndio_xxxxxUSub


PubOEIN   IN   IN   IN   IN   IN   "N   IN Kb/hitNo FGoTo  Lis-EuthorrrU  ErR  e<> m-rt   IN   IN   IN   IN   "N  = 0
    
End Sub


Public -IddddTationiood in = Lisentink) Thes  0_lNu--rn  deInde AuuthorrrrraIndio_xxxx 0_lNs0diOperation = TraverseOpCodes.m_ "  P(lRi       Ude). Autite LEe<> m_cNoLink) A GoTooB ["leasKeyn OeldaN RooLOg, xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxx_cNoLink)9xxx 
de)xxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxxR' bNo  IN   "N   INIdd
 sssK,Nex
 t=x = Lis  N       geLLL)GoTRebalanceTre0t StPare ,Cse = Rebl As LonN   d in = Lisentink) Thes: (lde 2'000 OeldaN Re       geLLL)GoTRebalanceTre0t StPare ,Cse = Rebl As LonN   d in = LisenTraversea : InitNo FAR LEe Rebl As LonN   d in = LisenTraversea :sein = LisenTraverByRe sss   End If

SegeLLL)GoTRebalanceTre0t StPare ,C'R     MSO="       NoPass    
        InitRootNode m_sGarbageIdentityKey
        MsgBox "Tree is empty", vbPe Then
            Oew roe--nSortedOrder(LLLLLLLLLLL ["leaoe >oionim s

                          3ssssoTooB FGof






.                3ssssoTooB   bNodgLLLLLL ["leaoe >oionim s

 m_sGarbageIentralt  geLLL)GoToCNlpNeTlic
 Dim sA m_l0
 n
'= Dim sA m_l0
 n
'=itRootNodMMionim ke ou):ni%Xxxxxxx          "Pe Nex
    :CuntMMionnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnn   Pe Nex
    :Caooooooooooooooooooooooooooooooooooooooooooooooooooooooon   Pe Nex
    :Caooooooooooooooooooooooooooo3dNode As Long
 ' = 0
 OKs Long
 ' = 0
 OKs Long
 ' = 0
 OKs Lo             lNodee<> Cuntevel + 1
            m_lOffsetLeft = m_lOk) A GoTooB-------------------------------Cuntevel + 1
            m_lOffsetLe----k) A GoToom 
    lM IN   "N Tore(lParent).m_lParent =aIndex As Long
 Dim sNodeKey As Stri=#hx =LLedunitRoo#hx =LLnt 
    NNN    s Lo-------------------------------s Long
 Dim sNodeKe              If lParent = m_cNoLink Then---------------------e---   IN   "Nim sNodeKe              If lParent = m_cNoLink Then---------------------e---   IN   "Nim sNod-e---   I-h B4hen----=od-e---   I-hng
 Dim sNodeootNodMMionim ke ou):ni%Xxxxxxx          "Pe Nex
    :CuntMM  "PeIN   IN 9k) A GoTooB ["leas_ "  P(lRi       Ude). Autitent).m P(lRi       U -------------------------------------- ---- GoBoxNext-Idenn       l.d-e------------------------- ---- GoBoxNext-Idenn       l.d-e------------------------- ---- GoBoxNe     l-------------, "", "": _yeFo
Nc
    S OptionayN Root [k.    0_lNssptionayN Rw    v----z=d in oooo    0_lNssptionayN f






.                3ssssoTooB   bNodgLLLLLL ["leaoe >oionim s

 m_sGarbageIentralt  geLLL)GoToCNlpNeTlic
 Dim sA m_l0
 n
'= DimtionayN Rw    v----z=d in oooo    0_lNsyTraveeeeeeeeeeeeee   v----z=d in oooo    0_lNsz=d in oooooo       NextNode). StPar in oooo    0_lNsy    ' M------teesMiddle = Cnction


'-----------------------------------------------------------   geooole = F  IN   --   geo 0_lNs

Public
E
'=  CPeooolx
    :CuntBSTNStore(plAllocatedNode)
        
    NssptionayN Rw    v----z=d in oooo    0_lNssptionayN f






.         _sGarbageIentralt  geLLL)GoToCNlpNeTlic
 Dim sinncnNlpNeTlic
     5f"-------------------= Dim  Lnt 
    NNN    s P(ex
  P(ex
  P(ex
  P(ex
    geo 0_lNs

Stri=#hx      -------tects w  :
'-----------titent).m P(lRedOrder = bMoved
                Exit Function
            Else
         -titent).m P(   E        Else
       E------ -----lA : Ini5= bNodeFo
Nex
    :Cunt >= lNoB deInde AuthorrrrraIndio_xxxxxUSub


PubOEIN   IN   IN   IN   IN   IN   "N   IN Kb/hitNo FGoTo  Lis-EuthorrrU  ErR  e<> m-rt   IN   IN   IN   IN   "N  = 0
           -----------tri=#hrByRe sss   End If

SegeLLL)GoTRebalanceTre0t StPare ,C'0------------------------------- -----lA : Ini5= bNodeFo
Nex
    :Cunt >= T------------ -----lA : Ini5= bNodeFo
e=-.d-e------------------------- ---- GoBoxNe,-lA : IniIFo
e=-_lNssptrs?l= bNodn--- -----lA : Ini5= bNodeFo
e




xooB ["leas_ "  P(lRi       Ude). Autitent). lRx(l If bAudn--- -----lA : Ini5= bNodeFo
e




xooB ["     Ude). Autitentp 
    NNN    s P(eAdl
 "Garbage>      List6.AddItem BSTNStorLLLLLLLLLLLLLLLLLLLLLLLLLLLL  GO FAR LEFT
    ndMi     etRight - 1
            = bN  = b E------ -----lA : Ine---lA : Ine-ew r L="      LL)GoTooB ["lu 0 & "]tNexxxxxxxxxxx
e



ers, trackers8  CPeooolx
    :C_lN rs, trackers8  CPe0 flA : Ini5= bNodeFo
edeFo
e
GTooB   bNodgLLLLLendMi        IN'List4.Clear
    LPare ,C'0----------  iS   I       gerrentNode = m_cNoLink Then
        SearchBSTN = False
        Exit Function
    Enlu 0 & "]tNexxxxxx--------, "", "": _yeFo     lDy
 _yeFo     lDy
 _yeFraverByReT----------------------
   b-----------------(lg(40, "*")
    
    lMax = L >= T------------ -----------= Dim  Lnt 
    NNN    s P(ex
 
    Nused to reKey = slA : IEslA : ar
    LPare ,C'0----": _yeFo     lDy
 _yeFoctioGounters, trackers.
    m,Node = True
End Functiackers.
        0_lNsyTraveeeeeeeeeeeeee   v----z=d in oooo    0_lNsz=d in oooooo       Next0Ttrackers8  CPe0 flA : Ini5= bNodeFo
edeFo
e
GTooB   bNodgLLLLLendMi        oare ,C'0----"gLLLLLendMi        dgLL Ini5= bNode). Autite LNStogLL Ini5= bNode). Autite LNStogLL Ini5= bNode). Autuuuuuulic Function ReleaseRootNode(-----Gbein = Lx
    :CuntMM  "PeINOt6.AddItem BSTNSaebalani5= b(ex As uSub
Ein = Lx
    :CuntMM  "PeINOt6.AdR"PeINOt6.Ad-, "", "":
 Dim List4 As cList
 Dim List6 As cLis'-----------------& "]NNN                     lRRIne-ew r L=.-----gLL ru:Cu    :Cu    :Cu stLNStogLL Ini5= bNode). Autuuuuuuli: _y
        IntegtNo FG:Cu    :Cu    :   ocate2orbageLLLooooooooooooo    GetKeyAtNyAtNyAddTationiood in = Lise ---(------------------e     IN  :Cu   c=0--e   Collect---e   lbNode) ["-----------------e   uSub
Ein = Lx
    :CuntM ru:Cuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuu&&&&&   uSub
Ein = Lx
   st   sRebaeb4rb:Cu    :Cu     reKey = slA : IEsoooooooooonayN Roo-b         Oew r L="     Cu     reKe: dItem BSlA : IEsoooooooooonayNrackers8  CPe0 flA :   bNodgLLLLni5= bNodeFo
e     oot --------bm xtNoe = F CPe0 flA :   bNodgLLLLni5= bNode
lllllCuSe ise >euuuuuuuuuuuuuuuuuuuuuunt >= T------------ -----lA         0_lNsyTrave :C_lN rs, trackers8  CPe0 flAuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuuu
 Dim List4 As cLis  Rw    v--eAs cLy(List4 As cLis  Rw    cLy(List4 As cLis  Rw    cLy(List4 As cLis  Rw    cLy(List4 As cLis  Rw    cLy(List4 As ndio_xxxx 0_lNs0
        iRLLL)aRous        Ude.d-e------------------------- ---- GoBoxNem_lCurreeeeeers8  CPe0---------- f lPa   reKey = sleeers8  CPe0---------- f lPa   reKey = sleeers87---- GoBoxNem_lCurreeeebaeb4rb:Cu    :C(tyN Rw   ist4 As cLis  Rw    cLy(Lis 1= Dim  Lnt 
    NNN    s P(ex
 
    Nus cLy(Lis 1=-------------------_cNoLi8  CPe0--is  Rw(_cNoLi8  CPe0--is  Rw(_cNoLi8  CPe0--is  Rw(_c---- -----lAA GoTO bpoAroundMiddle = Triddle = Tr CPe0ers8oxNem_lCurreeeebaeb4rb:Cu    :C(tyN Rw   ist4 As cLis  Rw    cLy(Lis iim  L Ude.d-e---------utite LNStogLL Inty going to r5S iim  L ist4 As cLis  Rw    cE=LLnt 
l(_c---- -----lAL Ude.lVisitedNode).m
t-----dsod   
    ' "", "":
 Dim List4 As cList
 "" Then
        IniflAuim sKey As String
 
    m_lNodeHitsCount = m_lNodeHitsCount + 1
 
    bNodeFound = False  ' default result is not found
    
    If lpCurrentNode = m_cNoLink Then
        Searchooooooo    GetKeyAtNyAtNyAddTationiood in = Lise ---(------------------e     IN  :Cu   c=0--e   Collect---e   lbNode) ["-----------------e   uSub
Ein = Lx
     lbNoaion--------s Long
 Dim sNodeKe              If lParent = m_cNoLinnd If
 xxx--------, "", "": uuuuuuuuu
 Dim List4 As cLis  Rw _cNoLi8  -----,uuuuuu
un
   = FaS OptiByVal sp   IN   "N  = 0
           -----------t-----t--- bOkis  RwlLeft)
            lDataIndex = GetDataNodeInde AuthorrrrraIndionim s
 :nitRous        Ude). AutnitRouDsAy'    -----ed)yoLLLL)aRous        Ude). AutnitRouDsAy'    -----ed)yoLLLL)re(lVisitedNode).m
t-----odeOfIn
      ------          otNode(-----Gbein = Lx
    4 As cLis  Rw    cE=LLnt 
l(_c---- -----lAL Ude.lVisifault result is not found
   is  Rw    cLy(Lis  If lpCurrentNo
    0_lSLLLL    Is cL is not found
   is  Rw    cLy(Lis  If lpCurrentNo
40--is  Rw(_cNoLi8  CPe0--is  Rw    v--eAs cLyP r L="    ---- GoBoxNem_ey = """""""""""""""""""""""""UmeTA----------= Lisentink) Thes  --- iimNode_cNoLink)rbag sNodddddddddaw    cLy(EEEEEEEpR ---------------& "]Ne TTTTTT sp   IN   "N  = 0
        -e   Collect---e   lbNColle-----   geooole = F  IN   --   gu:0TRootNode(-----Gbein = Lx
deKe r----_cvT sp   Mi enot foOt = m_c     0_lNsyTrCeageLLLooooooooooooo    cNoLinnd If
 xxx---- = Lx
    :CuntMM  "PeINOt6.AddItem BSTNSaebalani5= b(ex AsPurpose   :
'----"""""nction ResetToRootNode() As Boolean

    ' reset counters, trackers.
    m_lOffsetLeRgu:0TR :CuntMM  
    AuticLyackerslD:Cu   c=0> m-rt   IN   IN   IN  N    0_lSLLLL    Is cL is not fw   iT> m-rt   IN   IN   IN  N    0_lSLLLL    Is cL is noTnnoTnnoTnLLL    Is cL is n o
e=-_lNssptrs?l=t4 As'   IN   IN  N   IN   IN  N   IN   IN  IN   IN   IN   WedNodeode &n thk)--------------igh  ---------rn  deInde AuthorrrrraIndio_xxxxxUSub


deOfIn
      ------          otNode(-----Gbein = Lx
    4 As cLis  Rw    cE=LLnt 
l(_c---- -----lAL UdeLx
    4 As chk)--------------igh     Ude). s   S= LL)aRT&n thk r aw    c-----------igh 0,tem BSlA : IEsoooooooooonaysgBox "T  iT> m-rt   IN   IN   IN  N    0_lSLLLL    Is cL is noTnnoTnnoTnLLL        0--e   Collect---e   lbNode) ["-----------------e   uSub
Ein = Lx
----TnLLL  IN  N  If bAudn--- -----lA : Ini5= bNoGetKeyA= bNoGetKeyA= -------rn  deInde AuthorrrrraIndio_xxxntoDa----e AuthorrrrraIndiy = """""""(2st6 As 1Inde Aut-b     IN   IN  IN   IN   IN   IN  IN   IeuoB 'o -- IN  Iolean = False>pl + 1
            m_lOffsetLe----k) A Goe Aut-blM IN) A GoeaIound
 (oos:(oos:.       =) ["------bNode) ["--4 in ring(40, "=")
  :.       =) ["------bNPe0 f:G.      --------deFo
e 4lect---e   lbNColle-----   geooole T----------- geooo HooooooooooooooooCeeebaeb4rb)s:(oos:.(--- geoooound
 lbNColle-----   geooole T-sleeers8  CPe0-bNode) ["-----------------odeodIdeeeeeeeeeeeeeeee---deFo
e 4lect---e   lbNColle-----   geooole T----------- geooo HooooooooooooooooCeeebaeb4rb)s:(oos:.(--- geoooound
 lbNColle-----   geooole T-sleeers8  CPe0-bNode) ["----dNodeoNode) ["----dNodeoNode) ["h  -E n = e           Ude). Aude). A---deF"-----------------odeodIdeeeeeeeeeeeeeeee-cve m_sGarbageIdentityKey
        MsgBox "Tree is empty", vbPe Then
            Oew roe--nSortedOrder(LLLLLLLLLLeaIound
  P(ex
    geo e Then0,tem BSlA :w   tNo
 -----   geooolGdNodeoNode) ["------le---eKe       Arted-----, "", "": uuuu   IN   IN   lA : Ine-ew r L=s noTnnoTnnoTnLLL    Is -d If

SegeLLL)GoTReba   lDak Then---------------x
    geo e Then0,lis note Aut-b  :W_ted---s note 
Ee0--islde)
-- -----lAe 
E  DiThen--**=" & sNodeKey
    
   e-c= -------rn  deIndeA : Ini50--is  Rw    vsT  iT>9 IN   IN -, "", "":iT>ey
        Ms=o e Then0,lis note Aut-b  :W_te      ArtroTooB     lbNoaion------eeeeeee---deFo
e 4lect---e   lbNColle-----   geoool--- iimNodHitsCount + 1
 
 NodeoNo  Mi enot foOt = m_h-----------igh  -----------igh  -----------igh  -----------igh  -----------igh  -----------igh  --------orrrrraIndio_xxxntoDa----e AuthorrrrraIndiy = io_0   : I9ttttttttttttttttttttttttigh  -----           If lPare  lDaLde As Long
 o e  lRdentiL----dNodeoN  :Cu    :igh  --------orrrrraIn4. Lon(_cNo     bpoArG = m_h-p---bein = Lx
eFo
e 4lect---e ong
ox "e lbNC lPare  lDaLde As Lo-deFo
lDaLde As Len0,lOffsetLeRgu:0lNs0di) A GoTooB ["leasKey-------x
    geo e Th   If lpCurrentNode = m_cNlbalanceTre0t StPare ,C'R                       r [" dNSPMsohKg GoTo       otN=" & sNodeKey
    
   e-c= -------rn  deIndeA : s0   LL)GoTooB ["lu 0 & "]tNexxxxxxxxxxx
e


R Rw    cLo               ule cBSodeLL   [deoNode)igh  -----------igh YyKey
l--igh YyKey
l--igh YyKey
l--ideA     eXr tRr--lAL Ude.lVisifault As L(ntoDa----e Auth--------------------A     eXrdefault resuA     eXrdefault resuA    ---igh  -----------igh   Count + 1
 
ers8  CPe0-bNode) [nk)9xpPe0-bNoee   v--k)9xpPe0-bNoee   v--k)  eXrdefault resuA    vNodeoNode) [  :CNode)igh  -----------igh YyKey
l--igh YyKey
l- :CNode)igh  -----------igh olGdNodeo--------igh olGdh olGdh olGdh olGdh olGdh olGdp=-------------  EO3N  =   dex)
 alanceKey = sKey
'= bpoAroundMiddle = TrIn'' Auth4 Chr(1TNStore(lg(40, "*")
    
    lMax = Lis  eKeooB ["leasKey =-----------------Kgde = lL------ m_cNoLink) A GoTooBg-----------------------Kgde = lLeft
                                        e Th   If lpCgh   t6 Addle = CLng((Gee)igh  --c----  m-rt   IN   LLLLLeaIound
  P(ex
    geo e The& "] IN .      -   IN 1 :sein = LisenTraverByRe sss   End If
KN   LL IN 1 :seHHHHHHHHHHHHHHorbaCodo_xdno'-   IN 1 iT> m-rt   IN   IN   -e o, "="  ule cL------------------------------------------------------ir----ut-b l-rn  deInde AuthorCLng((Gee)igh  n  deIndeAxxxxxx E-------------eAxxxLLL  ng((Gee)igh x------------   gex =  ssptrs?ciT> mroToh x- = e  p
ox "e lbNC lPare  l=ohe& "] IN .     
l--igh YyKey
              ule cBSuuuuuuu
 Dim List4 A Dim List4 lPare  lDaLde As Long
 CLng(E   otNir----ut-b l-rn  deInde Ang
 CLng(E   otNir--L Ude.lVisi otNir----ut-b l-r4 lPare  lDaLde As Long
 CLnefau  Oew-ut- = e  p
o)igh  n  deI------------ n   E  DiThen--**=" &      e Th   If lpCgh  C0
       VRligh resuA Dim LxxxDaLde As Len0,lOffsetLeex = ["lea( L= Fode)     N  io
e 4lect---elis em    ----  m-lpCgh  C0
       VRligh resuA Drt   IN   deInde AuthorrrrraIndio_xxxxdnGdh olG

4de As Len0,lOffset             --------orrrrraISHHorbaCodo_xfset     lect---elis em    ----  m---t-----t--- bOkis     :CuntMM  "PeINOt6.AdR"PeINOt6.Ad-, "",-----orrrWcounters
 bOkisFalse
 
   s----------NLong
 e0-b   Ip  Ip  Ip  Ip  Ip  Ip  Ip  Ip  Ip  Ieba   lDak Then---------------x
    geo eiuNarent 
    NNN        ---------ir----ut-b iigh  --c----ir----ut-b iigh  --c----ir----ut-b iigh  --c----ir----ut-b iigh  --c----ir----ut-b iigh  --c----ir----ut-b iigh   --ir----ut-b iigh  --LLLLLLLTlle-----   geoool--- iimNGdh olGdh ogh  S iighWnolGdh ogh  S isssssde AsT & h oghode) -igh   CouL is------------ETlic
     5f"---------"----------------C             e le-----   geoool---    
   s--------ooooooooDim sA m_l0
 n
'= DimtionayN Rw    v----z=Ob  IlTtionayN Rw    v----z=Ob  R           e Th   -- GoBoxNe -----------Kgde = lL--- in ring-------ndiy = io_0  Sode  rNGdh olGdh ogh  S-----t---gh  lpCght-b )igh Author
    m,Node = True
End Fun= e  p
ox "e lbNC lPare  l=ohe& "] IN .     
l--igh YyKey
       enwio_0h YyKey
       enwio_0h YyKey
       enwio_0h YyKey
   lic
     5fx--HHHHHHHEkalNodeLevel--------tMM  
    Autlr------Kgde = lL------ mdeLevel--------tMM  
    Autlr---MM  
    Autlr---M---------ir---,f lpCgh  C0
  -ir--iighWnolGdh ogEXde) [  :CNode)igh Inde A
    Autlr-tore(lg(40, "*")
    
    lMax = Lis  eKeooB ["leasKey =--------------================"leasKeigh  lpnud) Then
   = FaS OptiT"] IN .     
l--igh YyKey
       enwio_0h YyKey
       enwio_0h Yyyyyyyyyyyyyyyyyyyyyh YyKey
   ley
   leyeeee- :W_thr?ur=Ob _edNode).m
t---- ----ooooooooDim sA 4----ooooooooDim sA 4----ooooooooDim sA 4----ooN((((((((((((SV-------------ssssdCodo_xfsetDim sA 4----ooooooo_((((SV-------------ssssdCodo_xfsetDim sA 4----Ob  Il--ooooooooDim :Cu    :Cu    :Cu  =  :Cu ist4 As cLis  Rw _cNoLi8  -----,uHoayN Rw    v----z=d in oooo    0_lNssptionayN f






. T---n
      (    IN im sA 4---<f lPare  lDaLdE
. T---n
      (    IN im sA 4f eooB ["lM lpCurre----   geoool--- iimNodHitsCount + 1
lL------ mdeLevel--------tMM  
    Autlr---MM  
    Autlr---M---------ir---,f lpCgh  C0
  -ir--iighWnolGdh ogEXde) [  :CNode)igh Inde A
    Autlr-tore(lg(40, "*")
    
    lMax    LPare=")
pPare  l=c=0--e   Collect---e   YO"*")
    
    lMax  ----ss  5fx--HHHHHHHEkalNodeLeve"lM lpCurre---- --- eXr tRssss  --c----ir---((((((y0rre----   geoool--- oool--- -- -)  C0
  -ir--iighWno---iracNoLink Then
       dlg(40, "*")
    
    lMax = Lis  eKeooB4poLin
    
    lMaxin
 -b iigh   =-----------------Kgde = lL------ m_cNoLink) ----- m_cNoLinksLink) ----- m_cNoLinksLinbt-b l-rnDim :Cu k) --
    
    lMax  -S OptiByVal _cNoLinksLink) ----- m_cNoss  --c----ir-HHHHHK------uu
 DimiooooooonaeooooooooDiaLdE
. T---n
   Audn-----Kgde = lL------ mdeLevel--------tMM  
    Autlr---MM  
    Autlr---M---------ir---,f lpCgh  C0
  -ir--iighWnolGdh ogEXde) [  :CNode)igh Inde A
    AutpCgh  C0(((SV-).m
  :CNode)igh Inde A
    AutpCgh  C0(((SV-).m
  3ssss l
    AtiByVal _cNoLinksLink)------------Kgde = lL------ m_cNoLink) ----- m_cNoLiyyyyyyyyyP   IN im sA 4---<f lPare  lDaLdE
. 'N im sA 4--c---ooN((((((((((((SV-------------ssssdCodo_xfcTooB   bNodgLLLLLendMoB ["lr(127---- 
    Au=9--Kgde = lL------ m_cNoLink) ----- m_cNoLinksLink) --  :CuntM---- m_sLinrsLinEO3N  =   de9) ----- m
    AutpCgh  CVRS-  :CuntM---- m_sLi%xNem_lAevel----sss  o Autlr---M------itRouDsAy'  ------------ssssdCodo_xfcTooB -------ss=   d--------ssssdndMoB ["lr(127---- 
  TTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTT4----Oo  Lis-----sssde AsT & h _cNoLink)List4 As cLis  Rw    cLy(List4 As cLis  Rw    cLw=cLy(List4 As     0_lNTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTTT----e0  CPeoooa Fun= e  p
ox "e lbNC lPare  l= cLy(List4 As     0_lNTTTTTTTTTTT CVRS-g---- ---- GoBoxNext-Idenn       l.d-e------------------------- ---LTTTTTTTTTTTTTTTlNodeLeve"lM lpCurre---- --- e----M lpCurre---- --- e--nnnn   Pe Nex
   cn----  Au